import apt_pkg, glob, os, collections
from urllib.request import urlopen
import sys
apt_pkg.config["dir"] = os.path.abspath("root")
if "armhf" in os.getcwd():
    apt_pkg.config["apt::architecture"] = "armhf"
apt_pkg.init()

log = open("global-ben.log", "w")
cache = apt_pkg.Cache(progress=None)
srcrecords = apt_pkg.SourceRecords()
pkgrecords = apt_pkg.PackageRecords(cache)

bin_to_src : dict[str, str] = {}
proposed : dict[str, set[str]] = {}
release : dict[str, set[str]] = {}
proposed_vers : dict[str, set[str]] = {}
release_vers : dict[str, set[str]] = {}
proposed_bins : set[str] = set()
release_bins : set[str] = set()
removed_bins : set[str] = set()
virtual_bins : set[str] = set()
virtual_bins : set[str] = set()

build_depends : dict[str, set[str]] = {}

tasks : dict[str, set[str]] = {}
comments = collections.defaultdict(set)

for file in glob.glob("root/var/lib/apt/lists/*Packages"):
    with apt_pkg.TagFile(file) as tags:
        for section in tags:
            try:
                tasks[section["Package"]] = set(s.strip() for s in section["Task"].split(","))
            except KeyError:
                continue

for proposed_sources in glob.glob("root/var/lib/apt/lists/*proposed*Sources"):
	for source in apt_pkg.TagFile(proposed_sources):
		proposed[source["Package"]] = set(p.strip() for p in source["Binary"].split(","))
		proposed_vers[source["Package"]] = source["Version"]
		proposed_bins.update(set(p.strip() for p in source["Binary"].split(",")))
		bins = set(p.strip() for p in source["Binary"].split(","))
		for b in bins:
			bin_to_src[b] = source["Package"]
            

		if apt_pkg.parse_src_depends("canary [%s]" % " ".join(a for a in source["Architecture"].split() if a != "all")):
			build_depends[source["Package"]] = set()
			for or_group in apt_pkg.parse_src_depends(source.get("Build-Depends", "")):
				for pkg, _, _ in or_group:
					build_depends[source["Package"]].add(pkg)
			for or_group in apt_pkg.parse_src_depends(source.get("Build-Depends-Arch", "")):
				for pkg, _, _ in or_group:
					build_depends[source["Package"]].add(pkg)
			for or_group in apt_pkg.parse_src_depends(source.get("Build-Depends-Indep", "")):
				for pkg, _, _ in or_group:
					build_depends[source["Package"]].add(pkg)

for release_sources in glob.glob("root/var/lib/apt/lists/*_noble_*Sources"):
	for source in apt_pkg.TagFile(release_sources):
		pkg = source["Package"]
		release[source["Package"]] = set(p.strip() for p in source["Binary"].split(","))
		bins = set(p.strip() for p in source["Binary"].split(","))
		release_vers[source["Package"]] = source["Version"]
		release_bins.update(bins)
		if pkg in proposed:
		    removed_bins.update(bins - proposed[pkg])
		for b in bins:
			bin_to_src[b] = source["Package"]

		if apt_pkg.parse_src_depends("canary [%s]" % " ".join(a for a in source["Architecture"].split() if a != "all")):
			if pkg not in build_depends:
				build_depends[source["Package"]] = set()
				for or_group in apt_pkg.parse_src_depends(source.get("Build-Depends", "")):
					for pkg, _, _ in or_group:
						build_depends[source["Package"]].add(pkg)
				for or_group in apt_pkg.parse_src_depends(source.get("Build-Depends-Arch", "")):
					for pkg, _, _ in or_group:
						build_depends[source["Package"]].add(pkg)
				for or_group in apt_pkg.parse_src_depends(source.get("Build-Depends-Indep", "")):
					for pkg, _, _ in or_group:
						build_depends[source["Package"]].add(pkg)

# Determine virtual transitions
have_binaries = {}
for pkg in cache.packages:
	for ver in pkg.version_list:
		pkgrecords.lookup(ver.file_list[0])
		srcpkg = pkgrecords.source_pkg or pkg.name
		srcver = pkgrecords.source_ver or ver.ver_str
		if srcpkg in proposed and proposed_vers[srcpkg] == srcver:
			have_binaries[srcpkg] = True
			for prvpkg, prvver, _ver in ver.provides_list:
				# We are the only provider of this package
				if prvpkg in cache and all(ver.parent_pkg.name == pkg.name for _, _, ver in cache[prvpkg].provides_list):
					print("#", srcpkg, srcver, "- adding virtual package", prvpkg, file=log)
					proposed[srcpkg].add(prvpkg)
					proposed_bins.add(prvpkg)
					virtual_bins.add(prvpkg)
					bin_to_src[prvpkg] = srcpkg
				else:
					print("#", srcpkg, srcver, "- ignoring virtual package", prvpkg, file=log)


for pkg in cache.packages:
	for ver in pkg.version_list:
		pkgrecords.lookup(ver.file_list[0])
		srcpkg = pkgrecords.source_pkg or pkg.name
		srcver = pkgrecords.source_ver or ver.ver_str
		if srcpkg in release_vers and srcpkg in proposed and have_binaries.get(srcpkg) and release_vers[srcpkg] == srcver:
			for prvpkg, prvver, _ver in ver.provides_list:
				# We are the only provider of this package
				if prvpkg in cache and all(ver.parent_pkg.name == pkg.name for _, _, ver in cache[prvpkg].provides_list):
					if prvpkg not in proposed[srcpkg]:
						print("#", srcpkg, srcver, "- removed virtual package", prvpkg, file=log)
						removed_bins.add(prvpkg)


t64_transitions = {}
with urlopen("https://people.canonical.com/~vorlon/armhf-time_t/ftp-override-map.txt") as url:
	for line in url:
		a, b = line.decode("utf-8").split()
		if a != b and b not in virtual_bins:
			t64_transitions[a] = b
			if b in bin_to_src or a not in bin_to_src:
				virtual_bins.add(a)
				removed_bins.add(a)

with open("global-ben.missing-t64.txt", "w") as out:
	print("# source old new", file=out)
	for a, b in sorted(t64_transitions.items(), key=lambda ab: bin_to_src.get(ab[0], ab[0])):
		if (a in release_bins or a in proposed_bins) and b not in release_bins and b not in proposed_bins and b not in virtual_bins:
			if bin_to_src[a] in ["binutils", "e2fsprogs", "mrpt", "python3.10", "python3.11", "pam"]:
				continue
			print("src:" + bin_to_src[a], a, b, file=out)

transitions = {}
reverse_transitions = {}
rebuilds = {}

for srcpkg, builddeps in build_depends.items():
	removed_deps = set()
	for depname in sorted(builddeps):
		deptype = "Build-Depends"
		if depname not in removed_bins:
			continue
		if depname not in transitions:
			transitions[depname] = dict()
		if deptype not in transitions[depname]:
			transitions[depname][deptype] = set()
		print("# tracking", srcpkg, deptype, depname, file=log)
		transitions[depname][deptype].add("src:" + srcpkg)
		removed_deps.add(depname)

	if removed_deps:	
		try:
			reverse_transitions["src:" + srcpkg].update(removed_deps)
		except KeyError:
			reverse_transitions["src:" + srcpkg] = removed_deps


for package in sorted(cache.packages, key=lambda p: p.name):
	if not package.has_versions:
		continue
	ver = package.version_list[0]
	pkgrecords.lookup(ver.file_list[0])
	if package.name in removed_bins:
		print("# ignoring removed package", package.name, file=log)
		continue
	if package.name not in release_bins and package.name not in proposed_bins:
		print("# ignoring NBS package", package.name, file=log)
		continue
	should_rebuild=True
	if package.name in proposed_bins:
		if not any(f.archive.endswith("-proposed") for f, o in ver.file_list):
			print("# missing proposed build for", package.name, file=log)
			should_rebuild=False
		elif (pkgrecords.source_ver or ver.ver_str) != proposed_vers.get(pkgrecords.source_pkg or package.name, ""):
			print("# old proposed build for", package.name, file=log)
			should_rebuild=False
	elif (pkgrecords.source_ver or ver.ver_str) != release_vers.get(pkgrecords.source_pkg or package.name):
		print("# ignoring old release build for", package.name, "built from", (pkgrecords.source_ver or ver.ver_str), " - current:", release_vers.get(pkgrecords.source_pkg or package.name), file=log)
		continue
	removed_deps = set()
	for deptype, or_groups in ver.depends_list_str.items():
		if deptype in ["Conflicts", "Breaks", "Suggests", "Replaces", "Recommends", "Enhances"]:
			continue
		for or_group in or_groups:
			for depname, depop, depver in or_group:
				if depname in removed_bins:
					t64 = t64_transitions.get(depname, depname + "t64")
					if any(dn == t64 for dn, do, dv in or_group):
						print("# ignoring", package.name, deptype, depname, "as it is in an or group with", t64, file=log)
						continue

					#print("#", ver, deptype, depname, file=out)
					if depname not in transitions:
						transitions[depname] = dict()
					if deptype not in transitions[depname]:
						transitions[depname][deptype] = set()
					if should_rebuild:
						print("# should rebuild ", package.name, ver.ver_str, "due to", depname, file=log)
						transitions[depname][deptype].add(pkgrecords.source_pkg or package.name + ("*" if package.name in tasks else ""))
					else:
						print("# transition ", package.name, ver.ver_str, "due to", depname, file=log)
						transitions[depname][deptype].add("[" + (pkgrecords.source_pkg or package.name) + "]")
					removed_deps.add(depname)
		#if removed_deps:
		#	print(package.name, deptype, ", ".join(sorted(removed_deps)))

	if removed_deps:	
		name = pkgrecords.source_pkg or package.name
		if not should_rebuild:
			name = "[%s]" % name
		try:
			reverse_transitions[name].update(removed_deps)
		except KeyError:
			reverse_transitions[name] = removed_deps

	rebuild_for = set()
	rebuild_from = set()
	for transition in sorted(removed_deps):
		t64 = t64_transitions.get(transition, transition + "t64")
		if (t64 in proposed_bins or t64 in release_bins) and (t64 in virtual_bins or (t64 in cache and cache[t64].version_list)):
			rebuild_from.add(transition)
			rebuild_for.add(t64)

	if rebuild_for and should_rebuild:
		print("#", "Rebuilding", pkgrecords.source_pkg or package.name, "due to", package.name, ver.ver_str, "depending on", rebuild_from, "- should be", rebuild_for, file=log)
		try:
			rebuilds[pkgrecords.source_pkg or package.name].update(rebuild_for)
		except KeyError:
			rebuilds[pkgrecords.source_pkg or package.name] = rebuild_for

		for task in tasks.get(package.name, []):
			comments[pkgrecords.source_pkg or package.name].add(f"task:{task}")

out = open("global-ben.rebuild-for.txt", "w")
print("#!/bin/sh", file=out)
print("# log has moved to global-ben.log", file=out)
print("# syntax: rebuild-for \"<reason>\" <package to rebuild>", file=out)
for source, rebuild_for in sorted(rebuilds.items()):
    reason = " # {}".format(" ".join(comments[source])) if source in comments else ""
    print('rebuild-for "%s" %s%s' % (", ".join(sorted(rebuild_for)), source, reason), file=out)
out.close()

with open("global-ben.reverse.txt", "w") as out:
	print("# package: list of transitions it is involved/removed binaries it depends on", file=out)
	print("# [] mean that we inspected an old build, e.g. not yet built in proposed (or removed from release)", file=out)
	print(file=out)
	for package, blocking in sorted(reverse_transitions.items(), key=lambda pb: (-len(pb[1]), pb[0])):
		print(package + ":",  " ".join(sorted(blocking)), file=out)
		print(file=out)

print("# packages listed in [] are failing to build in proposed")
for transition, subtransitions in sorted(transitions.items()):
	for deptype, rdeps in subtransitions.items():
		t64 = t64_transitions.get(transition, transition + "t64")
		if (t64 in proposed_bins or t64 in release_bins) and ((t64 in cache and cache[t64].version_list) or t64 in virtual_bins):
			transition += " (rebuild-for " + t64 + ")" 
		print(transition, "Reverse-" + deptype + ":", " ".join(sorted(rdeps, key=lambda r: ("[" in r, r))))

	print()
